/*
 * X11 Keyboard Utility Library
 *
 * (c) Ken Lynch 2001
 *
 * Distributed under the terms of the GPL
 *
 * See COPYING for more details
 *
 */

#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "keyboard.h"

void parseKeyString(Display * dpy, MyKey * key, char *str)
{
    char *k;

#ifdef DEBUG
    printf("entering parseKeyString\n");
    printf("key string=%s\n", str);
#endif

    key->keycode = 0;
    key->modifier = 0;

    if(!strcmp(str, "none"))
        return;

    k = strrchr(str, '+');
    if(k)
    {
        key->keycode = XKeysymToKeycode(dpy, XStringToKeysym(k + 1));
        if(strstr(str, "Shift"))
            key->modifier = key->modifier | ShiftMask;
        if(strstr(str, "Control"))
            key->modifier = key->modifier | ControlMask;
        if(strstr(str, "Mod1"))
            key->modifier = key->modifier | Mod1Mask;
        if(strstr(str, "Mod2"))
            key->modifier = key->modifier | Mod2Mask;
        if(strstr(str, "Mod3"))
            key->modifier = key->modifier | Mod3Mask;
        if(strstr(str, "Mod4"))
            key->modifier = key->modifier | Mod4Mask;
        if(strstr(str, "Mod5"))
            key->modifier = key->modifier | Mod5Mask;
    }
}

void grabKey(Display * dpy, MyKey * key, Window w)
{
#ifdef DEBUG
    printf("entering grabKey\n");
#endif

    if(key->keycode)
    {
        XGrabKey(dpy, key->keycode, key->modifier, w, False, GrabModeAsync,
                 GrabModeAsync);
        XGrabKey(dpy, key->keycode, key->modifier | LockMask, w, False,
                 GrabModeAsync, GrabModeAsync);
    }
}

void ungrabKeys(Display * dpy, Window w)
{
#ifdef DEBUG
    printf("entering ungrabKeys\n");
#endif

    XUngrabKey(dpy, AnyKey, AnyModifier, w);
}
