/*
 * Oroborus Window Manager
 *
 * (c) Ken Lynch 2001
 *
 * Distributed under the terms of the GPL
 *
 * See COPYING for more details
 *
 */

#include "oroborus.h"

char *progname;
Display *dpy;
Window root, gnome_win;
Colormap cmap;
int screen;
int depth;
long workspace;
CARD32 margins[4];
int quit = False, reload = False;
int shape, shape_event;
Cursor resize_cursor[4], move_cursor, root_cursor;

Client *clients = NULL;
CARD32 *client_list = NULL;
unsigned int client_count = 0;

MyKey keys[KEY_COUNT];
MyFont title_font;
int title_alignment, full_width_title, button_spacing, button_offset;
GC title_gc[2], box_gc;
MyPixmap sides[3][2], corners[4][2], buttons[4][3], title[5][2];
char button_layout[6];
int double_click_action;
int box_move, box_resize;
int click_to_focus, focus_new, raise_on_focus, raise_delay;
int window_placement, snap_to_border, snap_width;
int workspace_count = -1, wrap_workspaces;

int handleXError(Display * dpy, XErrorEvent * err)
{
    switch (err->error_code)
    {
        case BadAccess:
            if(err->resourceid == root)
            {
                fprintf(stderr, "%s: Another window manager is running\n",
                        progname);
                exit(1);
            }
    }
    return 0;
}

void cleanUp()
{
    int i;

#ifdef DEBUG
    printf("entering cleanUp\n");
#endif

    clientUnframeAll();
    unloadSettings();
    XFreeCursor(dpy, root_cursor);
    XFreeCursor(dpy, move_cursor);
    for(i = 0; i < 4; i++)
        XFreeCursor(dpy, resize_cursor[i]);
    XSetInputFocus(dpy, root, RevertToPointerRoot, CurrentTime);
    XCloseDisplay(dpy);
}

void handleSignal(int sig)
{
#ifdef DEBUG
    printf("entering handleSignal\n");
#endif

    switch (sig)
    {
        case SIGINT:
        case SIGTERM:
            quit = True;
            break;
        case SIGSEGV:
            fprintf(stderr, "%s: Segmentation fault\n", progname);
            cleanUp();
            exit(1);
            break;
        case SIGHUP:
            reload = True;
            break;
    }
}

void initialize(int argc, char **argv)
{
    struct sigaction act;
    char *locale;
    int dummy;

#ifdef DEBUG
    printf("entering initialize\n");
#endif

    progname = argv[0];

    printf
        ("\n\nOroborus (c) 2001 Ken Lynch\nDistributed under the terms and conditions of the GPL\n\n");

    locale = setlocale(LC_ALL, "");
    if(!locale || !strcmp(locale, "C") || !strcmp(locale, "POSIX") ||
       !XSupportsLocale())
        use_fontset = False;
    else
        use_fontset = True;

    dpy = XOpenDisplay(NULL);
    if(!dpy)
    {
        fprintf(stderr, "%s: failed to open display\n", progname);
        exit(1);
    }

    act.sa_handler = handleSignal;
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);
    sigaction(SIGTERM, &act, NULL);
    sigaction(SIGHUP, &act, NULL);
    sigaction(SIGSEGV, &act, NULL);

    root = XDefaultRootWindow(dpy);
    screen = XDefaultScreen(dpy);
    depth = DefaultDepth(dpy, screen);
    cmap = DefaultColormap(dpy, screen);

    XSetErrorHandler(handleXError);
    XSelectInput(dpy, root,
                 SubstructureNotifyMask | SubstructureRedirectMask |
                 ButtonPressMask | ButtonReleaseMask | FocusChangeMask |
                 PropertyChangeMask);

    shape = XShapeQueryExtension(dpy, &shape_event, &dummy);

    initICCCMHints(dpy);
    initMotifHints(dpy);
    initGnomeHints(dpy);
    gnome_win = XCreateSimpleWindow(dpy, root, -1000, -1000, 5, 5, 0, 0, 0);
    setGnomeHint(dpy, root, win_supporting_wm_check, gnome_win);
    setGnomeHint(dpy, gnome_win, win_supporting_wm_check, gnome_win);
    setGnomeHint(dpy, root, win_desktop_button_proxy, gnome_win);
    setGnomeHint(dpy, gnome_win, win_desktop_button_proxy, gnome_win);
    getGnomeHint(dpy, root, win_workspace, &workspace);
    getGnomeDesktopMargins(dpy, margins);

    root_cursor = XCreateFontCursor(dpy, XC_left_ptr);
    move_cursor = XCreateFontCursor(dpy, XC_fleur);
    resize_cursor[CORNER_TOP_LEFT] = XCreateFontCursor(dpy, XC_top_left_corner);
    resize_cursor[CORNER_TOP_RIGHT] =
        XCreateFontCursor(dpy, XC_top_right_corner);
    resize_cursor[CORNER_BOTTOM_LEFT] =
        XCreateFontCursor(dpy, XC_bottom_left_corner);
    resize_cursor[CORNER_BOTTOM_RIGHT] =
        XCreateFontCursor(dpy, XC_bottom_right_corner);
    XDefineCursor(dpy, root, root_cursor);

    loadSettings();
    clientFrameAll();

    XMapWindow(dpy, gnome_win);
    XSetInputFocus(dpy, gnome_win, RevertToNone, CurrentTime);
}

int main(int argc, char **argv)
{
    initialize(argc, argv);
    eventLoop();
    cleanUp();
    return 0;
}
