/*
 * Rc File Parsing Library
 *
 * (c) Ken Lynch 2001
 *
 * Distributed under the terms of the GPL
 *
 * See COPYING for more details
 *
 */

#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "parserc.h"

int parseRc(char *file, char *dir, char *rc[][2])
{
    char filename[512], buf[512], *lvalue, *rvalue;
    FILE *fp;
    int i;

#ifdef DEBUG
    printf("entering parseRc\n");
#endif

    if(dir)
        snprintf(filename, sizeof(filename), "%s/%s", dir, file);
    else
        strncpy(filename, file, sizeof(filename));
    fp = fopen(filename, "r");
    if(!fp)
        return False;

    while(fgets(buf, sizeof(buf), fp))
    {
        lvalue = strtok(buf, "=");
        if(lvalue)
        {
            for(i = 0; rc[i][OPTION]; i++)
            {
                if(!strcmp(lvalue, rc[i][OPTION]))
                {
                    rvalue = strtok(NULL, "\n");
                    if(rvalue)
                    {
                        if(rc[i][VALUE])
                            free(rc[i][VALUE]);
                        rc[i][VALUE] = strdup(rvalue);
#ifdef DEBUG
                        printf("%s=%s\n", rc[i][OPTION], rc[i][VALUE]);
#endif
                    }
                }
            }
        }
    }
    fclose(fp);
    return True;
}

int checkRc(char *rc[][2])
{
    int i, rval = True;

#ifdef DEBUG
    printf("entering checkRc\n");
#endif

    for(i = 0; rc[i][OPTION]; i++)
    {
        if(!rc[i][VALUE])
        {
            fprintf(stderr, "missing value for option %s\n", rc[i][OPTION]);
            rval = False;
        }
    }
    return rval;
}

char *getValue(char *option, char *rc[][2])
{
    int i;

#ifdef DEBUG
    printf("entering getValue\n");
#endif

    for(i = 0; rc[i][OPTION]; i++)
        if(!strcmp(option, rc[i][OPTION]))
            break;
    return rc[i][VALUE];
}

void freeRc(char *rc[][2])
{
    int i;

#ifdef DEBUG
    printf("entering freeRc\n");
#endif

    for(i = 0; rc[i][OPTION]; i++)
        free(rc[i][VALUE]);
}
