/*
 * Pixmap Utility Library
 *
 * (c) Ken Lynch 2001
 *
 * Distributed under the terms of the GPL
 *
 * See COPYING for more details
 *
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include "pixmap.h"

int loadPixmap(Display * dpy, MyPixmap * pm, char *dir, char *file,
               XpmColorSymbol * cs, int n)
{
    char filename[512];
    XpmAttributes attr;

#ifdef DEBUG
    printf("entering loadPixmap\n");
#endif

    pm->pixmap = None;
    pm->mask = None;
    pm->width = 1;
    pm->height = 1;
    snprintf(filename, sizeof(filename), "%s/%s", dir, file);
    attr.colorsymbols = cs;
    attr.numsymbols = n;
    attr.valuemask = XpmSize;
    if(n > 0 && cs)
        attr.valuemask = attr.valuemask | XpmColorSymbols;
    if(XpmReadFileToPixmap
       (dpy, XDefaultRootWindow(dpy), filename, &pm->pixmap, &pm->mask, &attr))
        return False;
    pm->width = attr.width;
    pm->height = attr.height;
    XpmFreeAttributes(&attr);
    return True;
}

void freePixmap(Display * dpy, MyPixmap * pm)
{
#ifdef DEBUG
    printf("entering freePixmap\n");
#endif

    if(pm->pixmap != None)
        XFreePixmap(dpy, pm->pixmap);
    if(pm->mask != None)
        XFreePixmap(dpy, pm->mask);
}

void scalePixmap(Display * dpy, MyPixmap * src, MyPixmap * dst, int width,
                 int height)
{
    XpmImage xi_src, xi_dst;
    int x, y, sx, sy, *src_data, *dst_data;

#ifdef DEBUG
    printf("entering scalePixmap\n");
#endif

    if(width < 1)
        width = 1;
    if(height < 1)
        height = 1;

    XpmCreateXpmImageFromPixmap(dpy, src->pixmap, src->mask, &xi_src, NULL);
    dst->width = width;
    dst->height = height;
    xi_dst.width = width;
    xi_dst.height = height;
    xi_dst.cpp = xi_src.cpp;
    xi_dst.ncolors = xi_src.ncolors;
    xi_dst.colorTable = xi_src.colorTable;
    xi_dst.data = malloc(sizeof(int) * (xi_dst.width * xi_dst.height));
    dst_data = xi_dst.data;
    src_data = xi_src.data;
    for(y = 0; y < xi_dst.height; y++)
    {
        dst_data = xi_dst.data + (y * xi_dst.width);
        for(x = 0; x < xi_dst.width; x++)
        {
            sx = (x * xi_src.width) / xi_dst.width;
            sy = (y * xi_src.height) / xi_dst.height;
            *dst_data = *(src_data + sx + (sy * xi_src.width));
            dst_data++;
        }
    }
    XpmCreatePixmapFromXpmImage(dpy, DefaultRootWindow(dpy), &xi_dst,
                                &dst->pixmap, &dst->mask, NULL);
    free(xi_dst.data);
    XpmFreeXpmImage(&xi_src);
}
